#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################

from helper_code import *
from my_preprocessing import *
from my_model import *

import os
import sys
import joblib
import numpy as np
import pandas as pd
import scipy
from scipy.stats import skew, kurtosis
from sklearn.model_selection import train_test_split
from sklearn.metrics import f1_score, confusion_matrix
from sklearn.model_selection import StratifiedKFold

from sklearn.ensemble import RandomForestClassifier
from xgboost import XGBClassifier
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import make_scorer

import tensorflow as tf
from keras.callbacks import ModelCheckpoint, EarlyStopping
from keras.models import load_model

################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################

# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')

    murmur_classes = ['Present', 'Unknown', 'Absent']
    num_murmur_classes = len(murmur_classes)
    outcome_classes = ['Abnormal', 'Normal']
    num_outcome_classes = len(outcome_classes)
    ########################################################################### My Custom Processing
    wave_info = list()
    wave_info2 = list()
    demo_info = list()
    murmurs = list()
    outcomes = list()
    # Get Data
    for i in range(num_patient_files):
        if verbose >= 2:
            print('    {}/{}...'.format(i+1, num_patient_files))

        # Load the current patient data and recordings.
        current_patient_data = load_patient_data(patient_files[i])
        current_recordings = custom_load_recordings(data_folder, current_patient_data)
        current_recordings2 = custom_load_recordings(data_folder, current_patient_data)
        current_features = custom_get_features(current_patient_data)

        # Extract features.
        wave_info.append(current_recordings)
        wave_info2.append(current_recordings2)
        demo_info.append(current_features)

        # Extract labels and use one-hot encoding.
        current_murmur = np.zeros(num_murmur_classes, dtype=int)
        murmur = get_murmur(current_patient_data)
        if murmur in murmur_classes:
            j = murmur_classes.index(murmur)
            current_murmur[j] = 1
        murmurs.append(current_murmur)

        current_outcome = np.zeros(num_outcome_classes, dtype=int)
        outcome = get_outcome(current_patient_data)
        if outcome in outcome_classes:
            j = outcome_classes.index(outcome)
            current_outcome[j] = 1
        outcomes.append(current_outcome)
    
    murmurs = np.vstack(murmurs)
    outcomes = np.vstack(outcomes)
    demo_info = np.vstack(demo_info)
    
    # Get Segmentation data
    for i in range(0, len(wave_info)): # Use only segmentation part
        n_loc = len(wave_info[i])
        for j in range(0, n_loc):
            tsv = wave_info[i][j][1]
            tsv = tsv[tsv.seg != 0].reset_index(drop = True)
            start = int(tsv.start[0]*4000)
            end = int(tsv.end[len(tsv)-1]*4000)
            wave_info[i][j][0] = wave_info[i][j][0][start:end]
    
    # Get Data per location & label
    pos_pos = []
    pos_neg = []
    unk_pos = []
    unk_neg = []
    neg_pos = []
    neg_neg = []
    for i in range(0, len(murmurs)):
        temp_mur = murmurs[i]
        temp_ab = outcomes[i]
        if temp_mur[0] == 1:
            if temp_ab[0] == 1:
                pos_pos.append(i)
            if temp_ab[1] == 1:
                pos_neg.append(i)

        if temp_mur[1] == 1:
            if temp_ab[0] == 1:
                unk_pos.append(i)
            if temp_ab[1] == 1:
                unk_neg.append(i)

        if temp_mur[2] == 1:
            if temp_ab[0] == 1:
                neg_pos.append(i)
            if temp_ab[1] == 1:
                neg_neg.append(i)
    
    pos_train, pos_valid = train_test_split(pos_pos + pos_neg, test_size = 0.4, random_state = 42)
    neg_train, neg_valid = train_test_split(neg_neg + neg_pos, test_size = 0.4, random_state = 42)

    train = pos_train + neg_train
    valid = pos_valid + neg_valid
    test = valid + unk_pos + unk_neg
    
    train_AV_x = []
    train_PV_x = []
    train_TV_x = []
    train_MV_x = []
    train_AV_x_demo = []
    train_PV_x_demo = []
    train_TV_x_demo = []
    train_MV_x_demo = []
    train_AV_y = []
    train_PV_y = []
    train_TV_y = []
    train_MV_y = []

    for i in train:
        temp_info = wave_info[i]
        n_loc = len(temp_info)
        for j in range(0, n_loc):
            x = temp_info[j][0]
            loc = temp_info[j][2]
            y = temp_info[j][3]
            if loc == 'AV':
                train_AV_x.append(x)
                train_AV_y.append(y)
                train_AV_x_demo.append(demo_info[i])
            if loc == 'PV':
                train_PV_x.append(x)
                train_PV_y.append(y)
                train_PV_x_demo.append(demo_info[i])
            if loc == 'TV':
                train_TV_x.append(x)
                train_TV_y.append(y)
                train_TV_x_demo.append(demo_info[i])
            if loc == 'MV':
                train_MV_x.append(x)
                train_MV_y.append(y)
                train_MV_x_demo.append(demo_info[i])
    
    train_x = train_AV_x + train_PV_x + train_TV_x + train_MV_x
    train_x_demo = train_AV_x_demo + train_PV_x_demo + train_TV_x_demo + train_MV_x_demo
    train_y = train_AV_y + train_PV_y + train_TV_y + train_MV_y
    
    valid_AV_x = []
    valid_PV_x = []
    valid_TV_x = []
    valid_MV_x = []
    valid_AV_x_demo = []
    valid_PV_x_demo = []
    valid_TV_x_demo = []
    valid_MV_x_demo = []
    valid_AV_y = []
    valid_PV_y = []
    valid_TV_y = []
    valid_MV_y = []

    for i in valid:
        temp_info = wave_info[i]
        n_loc = len(temp_info)
        for j in range(0, n_loc):
            x = temp_info[j][0]
            loc = temp_info[j][2]
            y = temp_info[j][3]
            if loc == 'AV':
                valid_AV_x.append(x)
                valid_AV_y.append(y)
                valid_AV_x_demo.append(demo_info[i])
            if loc == 'PV':
                valid_PV_x.append(x)
                valid_PV_y.append(y)
                valid_PV_x_demo.append(demo_info[i])
            if loc == 'TV':
                valid_TV_x.append(x)
                valid_TV_y.append(y)
                valid_TV_x_demo.append(demo_info[i])
            if loc == 'MV':
                valid_MV_x.append(x)
                valid_MV_y.append(y)
                valid_MV_x_demo.append(demo_info[i])
                
    valid_x = valid_AV_x + valid_PV_x + valid_TV_x + valid_MV_x
    valid_x_demo = valid_AV_x_demo + valid_PV_x_demo + valid_TV_x_demo + valid_MV_x_demo
    valid_y = valid_AV_y + valid_PV_y + valid_TV_y + valid_MV_y
    # Get Training-Validation Data
    train_seg_x, train_seg_y, train_demo = gen_inputs(train_x, train_y, train_x_demo)
    valid_seg_x, valid_seg_y, valid_demo = gen_inputs(valid_x, valid_y, valid_x_demo)
    # Random Shuffle for Training Data
    arr = np.arange(len(train_seg_x))
    np.random.seed(1)
    np.random.shuffle(arr)
    train_seg_x = train_seg_x[arr]
    train_seg_y = train_seg_y[arr]
    train_demo = train_demo[arr]
    ########################################################################### Custom Dataset End

    # Train the model.
    if verbose >= 1:
        print('Training model...')

    # Location Murmur Detection Model
    # set parameters
    n_filters = 128
    epochs = 50
    batch_size = 64
    patience = 10
    
    classifier_model = EEGNet(F1 = 32, F2 = 64, n_filters = n_filters)

    es = EarlyStopping(monitor = 'val_loss', mode = 'min', verbose = 1, patience = patience)
    mc = ModelCheckpoint(filepath = os.path.join(model_folder, 'classifier_model.h5'), monitor = 'val_rocauc', mode = 'max', save_best_only = True, verbose = 1)
    hist = classifier_model.fit([train_seg_x, train_demo], train_seg_y,
                     epochs = epochs,
                     batch_size = batch_size,
                     callbacks = [es, mc],
                     validation_data = ([valid_seg_x, valid_demo], valid_seg_y)
                     )
    # Load Deep Learning Model
    classifier_model = load_model(os.path.join(model_folder, 'classifier_model.h5'))
    
    # Find Prediction Threshold
    pred = classifier_model.predict([valid_seg_x, valid_demo])

    thv = []
    thv.append(threshold_detector(pred.flatten(), valid_seg_y))

    # Extract Decision Feature Rule
    if verbose >= 1:
        print('Decision Modeling...')
    df = pd.DataFrame(np.zeros((len(test), 18)), 
                      columns = ['AV_mean', 'AV_kurt', 'AV_skew', 'AV_ratio',
                                 'PV_mean', 'PV_kurt', 'PV_skew', 'PV_ratio',
                                 'TV_mean', 'TV_kurt', 'TV_skew', 'TV_ratio',
                                 'MV_mean', 'MV_kurt', 'MV_skew', 'MV_ratio',
                                 'murmur', 'abnormal'])

    for i in range(0, len(test)):
        temp_idx = test[i]
        temp_patient = wave_info2[temp_idx]
        number_loc = len(temp_patient)
        test_demo = demo_info[temp_idx]
        for j in range(number_loc):
            name_loc = temp_patient[j][2]
            if name_loc == 'Ph': # Except Phc
                continue
            temp_wave = temp_patient[j][0]
            seg_wave = []
            seg_demo = []
            for k in range(0, len(temp_wave) - 4000, 1000):
                seg_wave.append(get_wave_features(temp_wave[k: k+4000]))
                seg_demo.append(test_demo)
            seg_wave = np.array(seg_wave)
            seg_demo = np.array(seg_demo)
            if len(seg_wave) == 0: # if seg_wave == empty set, pass
                continue
            if name_loc == 'AV':
                pred = classifier_model.predict([seg_wave, seg_demo])
            if name_loc == 'PV':
                pred = classifier_model.predict([seg_wave, seg_demo])
            if name_loc == 'TV':
                pred = classifier_model.predict([seg_wave, seg_demo])
            if name_loc == 'MV':
                pred = classifier_model.predict([seg_wave, seg_demo])

            pred_mean = np.mean(pred.flatten())
            pred_kurt = kurtosis(pred.flatten())
            pred_skew = skew(pred.flatten())
            pred_label = []
            pred_thv = 0
            if name_loc == 'AV':
                pred_thv = thv[0]
            if name_loc == 'PV':
                pred_thv = thv[0]
            if name_loc == 'TV':
                pred_thv = thv[0]
            if name_loc == 'MV':
                pred_thv = thv[0]
            for n in range(0, len(pred)):
                if pred[n] >= pred_thv:
                    pred_label.append(1)
                else:
                    pred_label.append(0)
            pred_ratio = 1 - (np.sum(pred_label)/len(pred_label))

            df[name_loc+'_mean'][i] = pred_mean
            df[name_loc+'_kurt'][i] = pred_kurt
            df[name_loc+'_skew'][i] = pred_skew
            df[name_loc+'_ratio'][i] = pred_ratio
        df['murmur'][i] = np.argmax(murmurs[temp_idx])
        df['abnormal'][i] = np.argmax(outcomes[temp_idx])
    
    df = df.sample(frac = 1., random_state = 2).reset_index(drop = True)
    decision_features = df[['AV_mean', 'AV_kurt', 'AV_skew', 'AV_ratio',
                            'PV_mean', 'PV_kurt', 'PV_skew', 'PV_ratio',
                            'TV_mean', 'TV_kurt', 'TV_skew', 'TV_ratio',
                            'MV_mean', 'MV_kurt', 'MV_skew', 'MV_ratio']].values
    decision_target_murmur = df['murmur'].values
    decision_target_abnormal = df['abnormal'].values
    
    # Grid Search CV
    skf = StratifiedKFold(n_splits=5)
    parameters = {'class_weight':[None],
                  'max_depth':[4,5,6,7,8,9,None],
                  'n_estimators':[10,50,100,200,300,500]
                 }
    mur_clf = GridSearchCV(
        RandomForestClassifier(random_state = 0),
        parameters,
        scoring = make_scorer(w_acc1, greater_is_better=True),
        cv = skf)
    mur_clf.fit(decision_features, decision_target_murmur)
    mur_best_params = mur_clf.best_params_
    
    out_clf = GridSearchCV(
        RandomForestClassifier(random_state = 0),
        parameters,
        scoring = make_scorer(w_acc2, greater_is_better=True),
        cv = skf)
    out_clf.fit(decision_features, decision_target_abnormal)
    out_best_params = out_clf.best_params_
    
    # Training Decision Model
    murmur_classifier = RandomForestClassifier(n_estimators = mur_best_params['n_estimators'],
                                      class_weight = mur_best_params['class_weight'],
                                      max_depth = mur_best_params['max_depth'],
                                      random_state = 0)
    murmur_classifier.fit(decision_features, decision_target_murmur)
    
    outcome_classifier = RandomForestClassifier(n_estimators = out_best_params['n_estimators'],
                                      class_weight = out_best_params['class_weight'],
                                      max_depth = out_best_params['max_depth'],
                                      random_state = 0)
    outcome_classifier.fit(decision_features, decision_target_abnormal)

    # Save the model.
    save_challenge_model(model_folder, murmur_classes, murmur_classifier, outcome_classes, outcome_classifier, classifier_model, thv)

    if verbose >= 1:
        print('Done.')

# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):
    filename = os.path.join(model_folder, 'model.sav')
    return joblib.load(filename)

# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def run_challenge_model(model, data, recordings, verbose): # data = current patient data, recordings = current recording
    # Load Decision Model
    murmur_classes = model['murmur_classes']
    murmur_classifier = model['murmur_classifier']
    outcome_classes = model['outcome_classes']
    outcome_classifier = model['outcome_classifier']
    # Load Murmur Prediction Model
    classifier_model = model['classifier_model']
    thv = model['thv']
    # Segmentation Prediction -> Decision Rule
    wave, location = custom_get_features2(data, recordings) # wave list and location list
    demo = custom_get_features(data)
    df = pd.DataFrame(np.zeros((1, 16)), 
                      columns = ['AV_mean', 'AV_kurt', 'AV_skew', 'AV_ratio',
                                 'PV_mean', 'PV_kurt', 'PV_skew', 'PV_ratio',
                                 'TV_mean', 'TV_kurt', 'TV_skew', 'TV_ratio',
                                 'MV_mean', 'MV_kurt', 'MV_skew', 'MV_ratio'])
    
    number_loc = len(location)
    for j in range(number_loc):
        name_loc = location[j]
        if name_loc in 'Phc': # except Phc location
            continue
        temp_wave = wave[j]
        seg_wave = []
        seg_demo = []
        for k in range(0, len(temp_wave) - 4000, 1000):
            seg_wave.append(get_wave_features(temp_wave[k: k+4000]))
            seg_demo.append(demo)
        seg_wave = np.array(seg_wave)
        seg_demo = np.array(seg_demo)
        if len(seg_wave) == 0: # if seg_wave is empty set
            continue
        if name_loc == 'AV':
            pred = classifier_model.predict([seg_wave, seg_demo])
        if name_loc == 'PV':
            pred = classifier_model.predict([seg_wave, seg_demo])
        if name_loc == 'TV':
            pred = classifier_model.predict([seg_wave, seg_demo])
        if name_loc == 'MV':
            pred = classifier_model.predict([seg_wave, seg_demo])

        pred_mean = np.mean(pred.flatten())
        pred_kurt = kurtosis(pred.flatten())
        pred_skew = skew(pred.flatten())
        pred_label = []
        pred_thv = 0
        if name_loc == 'AV':
            pred_thv = thv[0]
        if name_loc == 'PV':
            pred_thv = thv[0]
        if name_loc == 'TV':
            pred_thv = thv[0]
        if name_loc == 'MV':
            pred_thv = thv[0]
        for n in range(0, len(pred)):
            if pred[n] >= pred_thv:
                pred_label.append(1)
            else:
                pred_label.append(0)
        pred_ratio = 1 - (np.sum(pred_label)/len(pred_label))

        df[name_loc+'_mean'][0] = pred_mean
        df[name_loc+'_kurt'][0] = pred_kurt
        df[name_loc+'_skew'][0] = pred_skew
        df[name_loc+'_ratio'][0] = pred_ratio
    
    features = df.values
    ####################################################################

    # Get classifier probabilities.
    murmur_probabilities = murmur_classifier.predict_proba(features)
    murmur_probabilities = np.asarray(murmur_probabilities, dtype=np.float32)
    outcome_probabilities = outcome_classifier.predict_proba(features)
    outcome_probabilities = np.asarray(outcome_probabilities, dtype=np.float32)

    # Choose label with highest probability.
    murmur_labels = np.zeros(len(murmur_classes), dtype=np.int_)
    idx = np.argmax(murmur_probabilities)
    murmur_labels[idx] = 1
    outcome_labels = np.zeros(len(outcome_classes), dtype=np.int_)
    idx = np.argmax(outcome_probabilities)
    outcome_labels[idx] = 1

    # Concatenate classes, labels, and probabilities.
    classes = murmur_classes + outcome_classes
    labels = np.concatenate((murmur_labels, outcome_labels))
    probabilities = np.concatenate((murmur_probabilities.flatten(), outcome_probabilities.flatten()))

    return classes, labels, probabilities

################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################

# Save your trained model.
def save_challenge_model(model_folder, murmur_classes, murmur_classifier, outcome_classes, outcome_classifier, classifier_model, thv):
    d = {'murmur_classes': murmur_classes, 'murmur_classifier': murmur_classifier, 'outcome_classes': outcome_classes, 'outcome_classifier': outcome_classifier,
        'classifier_model' : classifier_model, 'thv' : thv}
    filename = os.path.join(model_folder, 'model.sav')
    joblib.dump(d, filename, protocol=0)


###########################################################################################
### Custom Function

# Load recordings(custom ver).
def custom_load_recordings(data_folder, data, get_frequencies=False):
    num_locations = get_num_locations(data)
    recording_information = data.split('\n')[1:num_locations+1]
    
    recordings = list()
    frequencies = list()
    for i in range(num_locations):
        entries = recording_information[i].split(' ')
        ##############################
        loc_murmur = data.split('\n')[1+num_locations+5][9:]
        recording_file = entries[2]
        tsv_file = entries[3]
        recording_loc = entries[0][:2] # location 추가
        if loc_murmur == 'Present':
            if recording_loc in data.split('\n')[1+num_locations+6]:
                loc_murmur = 'Present'
            else:
                loc_murmur = 'Absent'
        ##############################
        filename = os.path.join(data_folder, recording_file)
        recording, frequency = load_wav_file(filename)
        segment = pd.read_csv(os.path.join(data_folder, tsv_file), sep = '\t', names = ['start', 'end', 'seg'])
        recordings.append([recording, segment, recording_loc, loc_murmur])
        frequencies.append(frequency)

    if get_frequencies:
        return recordings, frequencies
    else:
        return recordings

def gen_inputs(train_x, train_y, demo_features):
    features = []
    target = []
    demo_fea = []
    
    train_y_num = []
    for n in range(0, len(train_y)):
        if train_y[n] == 'Present':
            train_y_num.append(0)
        else:
            train_y_num.append(1)
    for i in range(0, len(train_x)):
        for j in range(0, len(train_x[i]) - 4000, 1000):
            features.append(get_wave_features(train_x[i][j: j+4000]))
            demo_fea.append(demo_features[i])
            target.append(train_y_num[i])
    
    features = np.array(features)
    target = np.array(target)
    demo_fea = np.array(demo_fea)
    return features, target, demo_fea

# Extract features from the data.
def custom_get_features2(data, recordings):

    # Extract recording locations and data. Identify when a location is present, and compute the mean, variance, and skewness of
    # each recording. If there are multiple recordings for one location, then extract features from the last recording.
    locations = get_locations(data)

    recording_locations = ['AV', 'MV', 'PV', 'TV']
    num_recording_locations = len(recording_locations)
    
    recording_features = []
    location_features = []
    
    num_locations = len(locations)
    num_recordings = len(recordings)
    if num_locations==num_recordings:
        for i in range(num_locations):
            for j in range(num_recording_locations):
                if compare_strings(locations[i], recording_locations[j]) and np.size(recordings[i])>0:
                    recording_features.append(recordings[i])
                    location_features.append(locations[i])
                    
    return recording_features, location_features

def threshold_detector(pred_array, true_array):
    cut = np.arange(0, 1, 0.05)
    score = []
    for thv in cut:
        pred_array_copy = pred_array.copy()
        for i in range(0, len(pred_array)):
            if pred_array_copy[i] > thv:
                pred_array_copy[i] = 1
            else:
                pred_array_copy[i] = 0
        score.append(f1_score(pred_array_copy, true_array, pos_label = 0))
    score = np.array(score)
    return cut[np.argmax(score)]

def w_acc1(test_y, pred):
    cm = confusion_matrix(pred, test_y)
    weights = np.array([[5, 3, 1], [5, 3, 1], [5, 3, 1]])
    B = weights * cm
    weighted_accuracy = np.trace(B) / np.sum(B)
    return weighted_accuracy

def w_acc2(test_y, pred):
    cm = confusion_matrix(pred, test_y)
    weights = np.array([[5, 1], [5, 1], [5, 1]])
    B = weights * cm
    weighted_accuracy = np.trace(B) / np.sum(B)
    return weighted_accuracy

def custom_get_features(data):
    # Extract the age group and replace with the (approximate) number of months for the middle of the age group.
    age_group = get_age(data)

    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 6 * 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15 * 12
    else:
        age = 6 * 12 # 6 * 12 # mode

    # Extract sex. Use one-hot encoding.
    sex = get_sex(data)
    sex_features = 0
    if compare_strings(sex, 'Female'):
        sex_features = 1
    
    # Extract height and weight.
    height = get_height(data)
    weight = get_weight(data)
    if np.isnan(height):
        height = 110 #110
    if np.isnan(weight):
        weight = 23 #23
    bmi = np.round(weight / ((height/100)**2) ,2)
    if bmi > 30:
        bmi = 30
    
    # Extract pregnancy status.
    is_pregnant = get_pregnancy_status(data)
    
    # Extract recording locations and data. Identify when a location is present, and compute the mean, variance, and skewness of
    # each recording. If there are multiple recordings for one location, then extract features from the last recording.

    age = age/(15*12)
    bmi = bmi/30
    
    features = np.hstack(([age], [bmi], [sex_features], [is_pregnant]))

    return np.asarray(features, dtype=np.float32)
